
"""
The Device Code grant type is used by browserless or input-constrained devices in the device
flow to exchange a previously obtained device code for an access token.

See also https://oauth.net/2/grant-types/device-code/
"""

import time
import typing as t
import requests
from urllib.parse import urlencode
from databind.core import datamodel
from databind.json import from_json
from .common import AccessTokenResponse, get_default_session, get_token_polling, raise_for_response


def get_device_code_parameters(client_id: str, client_secret: str, device_code: str) -> t.Dict[str, str]:
  return {
    'grant_type': 'urn:ietf:params:oauth:grant-type:device_code',
    'client_id': client_id,
    'client_secret': client_secret,
    'device_code': device_code,
  }


@datamodel
class DeviceCodeResponse:
  device_code: str
  user_code: str
  verification_url: str
  interval: int
  expires_in: t.Optional[int] = None


@datamodel
class Client:

  #: The URL where a device code can be requested.
  device_code_uri: str

  #: The URL that can be polled to convert the device code.
  token_uri: str

  #: The ID of the client.
  client_id: str

  #: The client secret.
  client_secret: str

  #: The scope of the device code.
  scope: t.Optional[str] = None

  session: t.Optional[requests.Session] = None

  def get_device_code(
    self,
    scope: t.Optional[str] = None,
    session: t.Optional[requests.Session] = None,
  ) -> DeviceCodeResponse:

    parameters = {'client_id': self.client_id}
    if scope or self.scope:
      parameters['scope'] = t.cast(str, scope or self.scope)

    session = session or self.session or get_default_session()
    response = session.post(self.device_code_uri, data=parameters)
    raise_for_response(response)
    return from_json(DeviceCodeResponse, response.json())

  def poll(
    self,
    info: DeviceCodeResponse,
    timeout: t.Optional[int] = None,
    session: t.Optional[requests.Session] = None,
  ) -> AccessTokenResponse:

    parameters = get_device_code_parameters(
      self.client_id,
      self.client_secret,
      info.device_code,
    )

    if timeout is None:
      if info.expires_in is not None:
        timeout = info.expires_in
    else:
      if info.expires_in is not None:
        timeout = min(timeout, info.expires_in)

    return get_token_polling(
      self.token_uri,
      parameters,
      info.interval,
      timeout,
      session or self.session,
    )




import abc
import typing as t
import sys
import time
import requests
from urllib.parse import parse_qsl, urlencode
from databind.core import datamodel
from databind.json import from_json
from . import __version__


class OAuth2Error(Exception):
  pass


class SlowDown(OAuth2Error):
  pass


class AuthorizationPending(OAuth2Error):
  pass


class RateLimitedExceeded(OAuth2Error):
  pass


class InvalidRequest(OAuth2Error):
  pass


class Timeout(OAuth2Error):
  pass


def raise_for_error_code(error_code: str) -> None:
  if error_code == 'authorization_pending':
    raise AuthorizationPending()
  elif error_code == 'slow_down':
    raise SlowDown()
  elif error_code == 'rate_limit_exceeded':
    raise RateLimitedExceeded()
  elif error_code == 'invalid_request':
    raise InvalidRequest()
  else:
    raise OAuth2Error(f'Unknown error: {error_code!r}')


def raise_for_response(response: requests.Response) -> None:
  if response.status_code // 100 not in (4, 5):
    return
  if 'application/json' not in response.headers.get('Content-Type', ''):
    response.raise_for_status()
  error_code = response.json().get('error')
  if not error_code:
    raise OAuth2Error('An unknown error occurred')
  raise_for_error_code(error_code)


@datamodel
class AccessTokenResponse:
  access_token: str
  token_type: str
  expires_in: t.Optional[int] = None
  refresh_token: t.Optional[str] = None
  scope: t.Optional[str] = None

  @staticmethod
  def from_response(response: requests.Response) -> 'AccessTokenResponse':
    raise_for_response(response)
    content_type = response.headers.get('Content-Type', '').partition(';')[0]
    if content_type == 'application/json':
      raw = response.json()
    elif content_type == 'application/x-www-form-urlencoded':
      raw = dict(parse_qsl(response.text))
    else:
      raise RuntimeError(f'unsupported Content-Type: {content_type!r}')
    return from_json(AccessTokenResponse, raw)


def get_user_agent() -> str:
  return f'Python-{sys.version.split(" ")[0]}/feedr.oauth2-{__version__}'


def get_default_session() -> requests.Session:
  session = requests.Session()
  session.headers['User-Agent'] = get_user_agent()
  return session


def get_token(
  token_url: str,
  params: t.Dict[str, str],
  session: t.Optional[requests.Session] = None,
) -> AccessTokenResponse:

  url = token_url + '?' + urlencode(params)
  session = session or get_default_session()
  response = session.post(url)
  return AccessTokenResponse.from_response(response)


def get_token_polling(
  token_url: str,
  params: t.Dict[str, str],
  interval: int,
  timeout: t.Optional[int],
  session: t.Optional[requests.Session] = None,
) -> AccessTokenResponse:

  session = session or get_default_session()
  start_time = time.perf_counter()
  end_time = start_time + timeout if timeout else None

  while end_time is None or time.perf_counter() < end_time:
    try:
      return get_token(token_url, params, session)
    except SlowDown:
      interval += 1  # TODO: Configurable back off strategy
    except AuthorizationPending:
      pass
    time.sleep(interval)

  raise Timeout(token_url, timeout)




import abc
import enum
import hashlib
import secrets
import string
import typing as t
import uuid
from urllib.parse import parse_qsl, urlencode

import requests
from databind.core import datamodel, field

__author__ = 'Niklas Rosenstein <rosensteinniklas@gmail.com>'
__version__ = '0.1.2'




"""
The Refresh Token grant type is used by clients to exchange a refresh token for an access token
when the access token has expired.

This allows clients to continue to have a valid access token without further interaction with
the user.

See also https://oauth.net/2/grant-types/refresh-token/
"""

import typing as t
import requests
from databind.core import datamodel
from .common import AccessTokenResponse, get_token


def get_refresh_token_parameters(
  refresh_token: str,
  scope: t.Optional[str],
  client_id: t.Optional[str] = None,
  client_secret: t.Optional[str] = None,
) -> t.Dict[str, str]:

  if (client_id or client_secret) and not (client_id and client_secret):
    raise ValueError('client_id and client_secret must both or neither be present')

  result = {
    'grant_type': 'refresh_token',
    'refresh_token': refresh_token,
  }
  if scope:
    result['scope'] = scope
  if client_id and client_secret:
    result.update({
      'client_id': client_id,
      'client_secret': client_secret,
    })
  return result


@datamodel
class Client:

  token_uri: str

  #: The refresh token previously issue to the client.
  refresh_token: str

  #: The requested scope must not include additional scopes that were not issued in the original
  #: access token. Typically this will not be included in the request, and if omitted, the service
  #: should issue an access token with the same scope as was previously issued.
  scope: t.Optional[str] = None

  #: The client ID and secret are optional if the client has not been granted a secret.
  client_id: t.Optional[str] = None
  client_secret: t.Optional[str] = None

  session: t.Optional[requests.Session] = None

  def get_token(
    self,
    scope: t.Optional[str] = None,
    session: t.Optional[requests.Session] = None,
  ) -> AccessTokenResponse:

    parameters = get_refresh_token_parameters(
      self.refresh_token,
      scope or self.scope,
      self.client_id,
      self.client_secret,
    )

    return get_token(self.token_uri, parameters, session or self.session)




"""
The Authorization Code grant type is used by confidential and public clients to exchange an
authorization code for an access token.

After the user returns to the client via the redirect URL, the application will get the
authorization code from the URL and use it to request an access token.

It is recommended that all clients use the PKCE extension with this flow as well to provide better
security.

See also https://oauth.net/2/grant-types/authorization-code/
"""

import enum
import hashlib
import string
import secrets
import typing as t
import uuid
import requests
from urllib.parse import urlencode
from databind.core import datamodel
from .common import AccessTokenResponse, get_token
from .refresh_token import Client as _RefreshTokenClient


@datamodel
class PKCE:
  """
  The Proof Key for Code Exchange (PKCE, pronounced pixie) extension describes a technique for
  public clients to mitigate the threat of having the authorization code intercepted. The
  technique involves the client first creating a secret, and then using that secret again when
  exchanging the authorization code for an access token. This way if the code is intercepted, it
  will not be useful since the token request relies on the initial secret.

  See also https://www.oauth.com/oauth2-servers/pkce/
  """

  class Method(enum.Enum):
    PLAIN = enum.auto()
    SHA256 = enum.auto()

  code_verifier: str
  code_challenge_method: Method

  @staticmethod
  def generate_code_verifier(length: int) -> str:
    if length < 43:
      raise ValueError(f'requested code_verifier length too small: {length}')
    if length > 128:
      raise ValueError(f'requested code_verifier length too large: {length}')
    pool = string.ascii_letters + string.digits + '-._~'
    return ''.join(secrets.choice(pool) for _i in range(length))

  @staticmethod
  def create(length: int = 128, method: Method = Method.SHA256) -> 'PKCE':
    return PKCE(PKCE.generate_code_verifier(length), method)


def get_authorization_request_parameters(
  client_id: str,
  state: str,
  scope: t.Optional[str] = None,
  redirect_uri: t.Optional[str] = None,
  pkce: t.Optional[PKCE] = None,
) -> t.Dict[str, str]:
  """
  Returns the URL query parameters for an OAuth2 user authorization request.
  """

  result = {
    'response_type': 'code',
    'client_id': client_id,
    'state': state,
  }
  if redirect_uri:
    result['redirect_uri'] = redirect_uri
  if scope:
    result['scope'] = scope
  if pkce:
    if pkce.code_challenge_method == PKCE.Method.PLAIN:
      result['code_challenge_method'] = 'plain'
      result['code_challenge'] = pkce.code_verifier
    elif pkce.code_challenge_method == PKCE.Method.SHA256:
      result['code_challenge_method'] = 'S256'
      result['code_challenge'] = hashlib.sha256(pkce.code_verifier.encode('ascii')).hexdigest()
    else:
      raise RuntimeError(f'invalid code_challenge_method: {pkce.code_challenge_method}')
  return result


def get_authorization_code_parameters(
  client_id: str,
  client_secret: str,
  code: str,
  redirect_uri: t.Optional[str] = None,
  pkce: t.Optional[PKCE] = None,
) -> t.Dict[str, str]:
  """
  Returns the URL query parameters for an OAuth2 authorization code request.
  """

  result = {
    'grant_type': 'authorization_code',
    'code': code,
    'client_id': client_id,
    'client_secret': client_secret,
  }
  if redirect_uri:
    result['redirect_uri'] = redirect_uri
  if pkce:
    result['code_verifier'] = pkce.code_verifier
  return result


@datamodel
class Request:
  auth_uri: str
  state: str
  pkce: t.Optional[PKCE]


@datamodel
class Client:
  """
  Represents an OAuth2 client that looks to authenticate with a server using the
  `authorization_code` grant type.
  """

  #: The URL of the server under which a user authorization request can be made. A user will be
  #: redirected to this page including certain parameters to authenticate.
  auth_uri: str

  #: The URL of the authorization server to exchange the authorization code with an access token.
  token_uri: str

  #: The ID of the client application.
  client_id: str

  #: The secret of the client application.
  client_secret: str

  #: The URL to which a user is supposed to be redirected to after the they authenticated with the
  #: authorization server. This URL will receive the authorization code. This can be overwritten
  #: for each individual #authorization_request().
  redirect_uri: t.Optional[str] = None

  #: The scope for the client's permissions with the user's access token. This can be overwritten
  #: for each individual #authorization_request().
  scope: t.Optional[str] = None

  #: Whether to use PKCS.
  use_pkce: bool = False

  #: The session to use when retrieving access token.
  session: t.Optional[requests.Session] = None

  def authorization_request(
    self,
    scope: t.Optional[str] = None,
    redirect_uri: t.Optional[str] = None,
  ) -> Request:
    """
    Creates an authorization request, returning a tuple of the URL that a user must be redirected
    to, the state ID of the request and the #PKCE data for the request. The state ID and PKCE data
    needs to be cached in a way that the #redirect_uri can retrieve the information based on the
    state ID.
    """

    state = str(uuid.uuid4())
    pkce = PKCE.create() if self.use_pkce else None
    parameters = get_authorization_request_parameters(
      self.client_id,
      state,
      self.scope or scope,
      redirect_uri or self.redirect_uri,
      pkce,
    )
    return Request(self.auth_uri + '?' + urlencode(parameters), state, pkce)

  def authorization_code(
    self,
    code: str,
    pkce: t.Optional[PKCE],
    redirect_uri: t.Optional[str] = None,
    session: t.Optional[requests.Session] = None,
  ) -> AccessTokenResponse:
    """
    Performa a POST request to the #token_uri to exchange the authorization *code* for an
    access token.
    """

    if self.use_pkce and not pkce:
      raise RuntimeError('use_pkce is enabled but no PKCE was passed')
    elif not self.use_pkce and pkce:
      raise RuntimeError('use_pkce is disabled but a PKCE was passed')

    parameters = get_authorization_code_parameters(
      self.client_id,
      self.client_secret,
      code,
      redirect_uri,
      pkce,
    )

    return get_token(self.token_uri, parameters, session or self.session)

  def refresh_token(
    self,
    refresh_token: str,
    scope: t.Optional[str] = None,
    session: t.Optional[requests.Session] = None,
  ) -> AccessTokenResponse:
    client = _RefreshTokenClient(
      self.token_uri,
      refresh_token,
      scope or self.scope,
      self.client_id,
      self.client_secret,
      session or self.session,
    )
    return client.get_token()




"""
The Client Credentials grant is used when applications request an access token to access their
own resources, not on behalf of a user.

See also https://www.oauth.com/oauth2-servers/access-tokens/client-credentials/
"""


import typing as t
import requests
from urllib.parse import urlencode
from databind.core import datamodel
from .common import AccessTokenResponse, get_token


def get_client_credentials_parameters(
  client_id: str,
  client_secret: str,
  scope: t.Optional[str] = None,
) -> t.Dict[str, str]:
  result = {
    'grant_type': 'client_credentials',
    'client_id': client_id,
    'client_secret': client_secret,
  }
  if scope:
    result['scope'] = scope
  return result


@datamodel
class Client:

  #: The URL on which a token can be exchanged.
  token_uri: str

  #: The client needs to authenticate themselves for this request. Typically the service will
  #: allow additional request parameters client_id and client_secret.
  client_id: str
  client_secret: str

  #: Your service can support different scopes for the client credentials grant. In practice, not
  #: many services actually support this.
  scope: t.Optional[str] = None

  session: t.Optional[requests.Session] = None

  def get_token(
    self,
    scope: t.Optional[str] = None,
    session: t.Optional[requests.Session] = None,
  ) -> AccessTokenResponse:

    parameters = get_client_credentials_parameters(
      self.client_id,
      self.client_secret,
      scope or self.scope,
    )

    return get_token(self.token_uri, parameters, session or self.session)
